@@ -31,33 +31,37 @@ module Data.Csv.Conversion
     , namedRecord
     ) where
 
-import Control.Applicative (Alternative, Applicative, (<*>), (<$>), (<|>),
-                            (<*), (*>), empty, pure)
+import Control.Applicative (Alternative, (<$>), (<|>), empty)
 import Control.Monad (MonadPlus, mplus, mzero)
-import Data.Attoparsec.Char8 (double)
-import qualified Data.Attoparsec.Char8 as A8
+import Data.Attoparsec.ByteString.Char8 (double)
+import qualified Data.Attoparsec.ByteString.Char8 as A8
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
 import qualified Data.ByteString.Lazy as L
 import qualified Data.HashMap.Lazy as HM
 import Data.Int (Int8, Int16, Int32, Int64)
 import qualified Data.Map as M
-import Data.Monoid (Monoid, mappend, mempty)
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
 import qualified Data.Text.Lazy as LT
 import qualified Data.Text.Lazy.Encoding as LT
-import Data.Traversable (traverse)
 import Data.Vector (Vector, (!))
 import qualified Data.Vector as V
 import qualified Data.Vector.Unboxed as U
-import Data.Word (Word, Word8, Word16, Word32, Word64)
+import Data.Word (Word8, Word16, Word32, Word64)
 import GHC.Float (double2Float)
 import Prelude hiding (lookup, takeWhile)
 
 import Data.Csv.Conversion.Internal
 import Data.Csv.Types
 
+#if !MIN_VERSION_base(4,8,0)
+import Control.Applicative (Applicative, (<*>), (<*), (*>), pure)
+import Data.Monoid (Monoid, mappend, mempty)
+import Data.Traversable (traverse)
+import Data.Word (Word)
+#endif
+
 #ifdef GENERICS
 import GHC.Generics
 import qualified Data.IntMap as IM
@@ -108,7 +112,7 @@ fromStrict = L.fromChunks . (:[])
 -- >         | otherwise     = mzero
 class FromRecord a where
     parseRecord :: Record -> Parser a
-  
+
 #ifdef GENERICS
     default parseRecord :: (Generic a, GFromRecord (Rep a)) => Record -> Parser a
     parseRecord r = to <$> gparseRecord r
@@ -253,6 +257,53 @@ instance (ToField a, ToField b, ToField c, ToField d, ToField e, ToField f,
         toField a, toField b, toField c, toField d, toField e, toField f,
         toField g]
 
+instance (FromField a, FromField b, FromField c, FromField d, FromField e,
+          FromField f, FromField g, FromField h) =>
+         FromRecord (a, b, c, d, e, f, g, h) where
+    parseRecord v
+        | n == 8    = (,,,,,,,) <$> unsafeIndex v 0
+                                <*> unsafeIndex v 1
+                                <*> unsafeIndex v 2
+                                <*> unsafeIndex v 3
+                                <*> unsafeIndex v 4
+                                <*> unsafeIndex v 5
+                                <*> unsafeIndex v 6
+                                <*> unsafeIndex v 7
+        | otherwise = lengthMismatch 8 v
+          where
+            n = V.length v
+
+instance (ToField a, ToField b, ToField c, ToField d, ToField e, ToField f,
+          ToField g, ToField h) =>
+         ToRecord (a, b, c, d, e, f, g, h) where
+    toRecord (a, b, c, d, e, f, g, h) = V.fromList [
+        toField a, toField b, toField c, toField d, toField e, toField f,
+        toField g, toField h]
+
+instance (FromField a, FromField b, FromField c, FromField d, FromField e,
+          FromField f, FromField g, FromField h, FromField i) =>
+         FromRecord (a, b, c, d, e, f, g, h, i) where
+    parseRecord v
+        | n == 9    = (,,,,,,,,) <$> unsafeIndex v 0
+                                 <*> unsafeIndex v 1
+                                 <*> unsafeIndex v 2
+                                 <*> unsafeIndex v 3
+                                 <*> unsafeIndex v 4
+                                 <*> unsafeIndex v 5
+                                 <*> unsafeIndex v 6
+                                 <*> unsafeIndex v 7
+                                 <*> unsafeIndex v 8
+        | otherwise = lengthMismatch 9 v
+          where
+            n = V.length v
+
+instance (ToField a, ToField b, ToField c, ToField d, ToField e, ToField f,
+          ToField g, ToField h, ToField i) =>
+         ToRecord (a, b, c, d, e, f, g, h, i) where
+    toRecord (a, b, c, d, e, f, g, h, i) = V.fromList [
+        toField a, toField b, toField c, toField d, toField e, toField f,
+        toField g, toField h, toField i]
+
 lengthMismatch :: Int -> Record -> Parser a
 lengthMismatch expected v =
     fail $ "cannot unpack array of length " ++
@@ -636,11 +687,18 @@ ws = A8.skipWhile (\c -> c == ' ' || c == '\t')
 parseOnly :: A8.Parser a -> B.ByteString -> Either String a
 parseOnly parser input = go (A8.parse parser input) where
   go (A8.Fail _ _ err) = Left err
-  go (A8.Partial f)    = go (f B.empty)
+  go (A8.Partial f)    = go2 (f B.empty)
   go (A8.Done leftover result)
     | B.null leftover = Right result
     | otherwise = Left ("incomplete field parse, leftover: "
                         ++ show (B.unpack leftover))
+
+  go2 (A8.Fail _ _ err) = Left err
+  go2 (A8.Partial _)    = error "parseOnly: impossible error!"
+  go2 (A8.Done leftover result)
+    | B.null leftover = Right result
+    | otherwise = Left ("incomplete field parse, leftover: "
+                        ++ show (B.unpack leftover))
 {-# INLINE parseOnly #-}
 
 typeError :: String -> B.ByteString -> Maybe String -> Parser a
@@ -797,9 +855,9 @@ class GFromRecord f where
     gparseRecord :: Record -> Parser (f p)
 
 instance GFromRecordSum f Record => GFromRecord (M1 i n f) where
-    gparseRecord v = 
+    gparseRecord v =
         case (IM.lookup n gparseRecordSum) of
-            Nothing -> lengthMismatch n v 
+            Nothing -> lengthMismatch n v
             Just p -> M1 <$> p v
       where
         n = V.length v
@@ -808,15 +866,15 @@ class GFromNamedRecord f where
     gparseNamedRecord :: NamedRecord -> Parser (f p)
 
 instance GFromRecordSum f NamedRecord => GFromNamedRecord (M1 i n f) where
-    gparseNamedRecord v = 
+    gparseNamedRecord v =
         foldr (\f p -> p <|> M1 <$> f v) empty (IM.elems gparseRecordSum)
 
 class GFromRecordSum f r where
     gparseRecordSum :: IM.IntMap (r -> Parser (f p))
 
 instance (GFromRecordSum a r, GFromRecordSum b r) => GFromRecordSum (a :+: b) r where
-    gparseRecordSum = 
-        IM.unionWith (\a b r -> a r <|> b r) 
+    gparseRecordSum =
+        IM.unionWith (\a b r -> a r <|> b r)
             (fmap (L1 <$>) <$> gparseRecordSum)
             (fmap (R1 <$>) <$> gparseRecordSum)
 
